/* fdformat.c  -  Low-level formats a floppy disk. */

/* Atari version speciality:
 * 
 * With the -t option, you can specify one or two twist factors. (If
 * there's only one, this one is used for both.) These are for track
 * twisting, i.e.  the increment of the offset of sector 1 to the
 * index mark. If this offset is choosen right, as little time as
 * possible is wasted after a track change until sector 1 is under the
 * head. The first twist factor is used between tracks on different
 * sides of the disk, but on the same cylinder. The second is applied
 * between tracks on different cylinders. In the last case, the FDC
 * has to do an additional step to the new track, so the delay between
 * end and start of reading is a bit more than without this stepping.
 *
 * As usual, the -n option suppresses verifying the disk.
 *
 * (Atari adaption by Roman.Hodek@informatik.uni-erlangen.de)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <linux/fd.h>
#include <asm/atafd.h>
#include <linux/major.h>
#include <linux/fs.h>


static int ctrl;
struct floppy_struct param;


#define SECTOR_SIZE 512
#define PERROR(msg) { perror(msg); exit(1); }


/* We cannot use FD_FILL_BYTE == 0xf6, since this writes sync marks instead of
 * data with the Atari FDC */
#define ATARI_FD_FILL_BYTE 0xe5


static void format_disk( char *name, int twist1, int twist2 )

{	struct atari_format_descr descr;
	int track, off;

	printf( "Formatting ... " );
    fflush( stdout );

    if (ioctl( ctrl, FDFMTBEG, NULL ) < 0)
		PERROR( "\nioctl(FDFMTBEG)" );

    for( track = 0, off = 0; track < param.track; track++ ) {
		descr.track = track;
		descr.head = 0;
		descr.sect_offset = off;
		off = (off + twist1) % param.sect;
		if (ioctl( ctrl, FDFMTTRK, (int)&descr ) < 0)
			PERROR( "\nioctl(FDFMTTRK)" );
		printf( "%3d\b\b\b", track );
		fflush( stdout );
		
		if (param.head == 2) {
			descr.head = 1;
			descr.sect_offset = off;
			off = (off + twist2) % param.sect;
			if (ioctl( ctrl, FDFMTTRK, (int)&descr ) < 0)
				PERROR( "\nioctl(FDFMTTRK)" );
		}
    }
	
    if (ioctl( ctrl, FDFMTEND, NULL) < 0)
		PERROR( "\nioctl(FDFMTEND)" );
    printf( "done\n" );
}


static void verify_disk( char *name )

{	unsigned char *data;
    int cyl_size, cyl, count;

    cyl_size = param.sect*param.head*512;
    if (!(data = (unsigned char *)malloc( cyl_size )))
		PERROR( "malloc" );

    printf( "Verifying ... " );
    fflush( stdout );
	
    for( cyl = 0; cyl < param.track; cyl++ ) {
		printf( "%3d\b\b\b", cyl );
		fflush( stdout );
		
		if (read( ctrl, data, cyl_size ) != cyl_size)
			PERROR( "read" );

		for( count = 0; count < cyl_size; count++ ) {
			if (data[count] != ATARI_FD_FILL_BYTE) {
				printf( "bad data in cyl %d\nContinuing ... ", cyl );
				fflush( stdout );
				break;
			}
		}
    }
    printf( "done\n" );
}


static void usage( char *name )

{	char *this;

    if ((this = strrchr( name, '/' )))
		name = this+1;
    fprintf( stderr, "usage: %s [-n] [-t twist[,twist2]] device\n",
			 name );
    exit( 1 );
}


int main( int argc, char **argv )

{	int verify;
    int twist1, twist2, user_twist1 = -1, user_twist2 = -1;
    char *name, *p, *p2;
	struct stat stbuf;
	
    name = argv[0];
    verify = 1;
    while( argc > 1 && argv[1][0] == '-' ) {
		if (argv[1][1] == 'n') {
			verify = 0;
			argc--; argv++;
		}
		else if (argv[1][1] == 't') {
			p = argv[1]+2;
			if (!*p) {
				if (argc < 2) {
					fprintf( stderr, "-t option missing argument\n" );
					usage( name );
				}
				argc--; argv++;
				p = argv[1];
			}
			user_twist1 = strtoul( p, &p2, 10 );
			if (p2 == p) {
				fprintf( stderr, "number expected after -t option\n" );
				usage( name );
			}
			if (*p2 == ',') {
				p = p2+1;
				user_twist2 = strtoul( p, &p2, 10 );
				if (p2 == p) {
					fprintf( stderr, "number expected after ',' in -t option\n" );
					usage( name );
				}
			}
			else
				user_twist2 = user_twist1;
			argc--; argv++;
		}
		else {
			usage( name );
		}
	}
	if (argc != 2)
		usage( name );
		
	if (stat( argv[1], &stbuf ))
		PERROR( "stat" );
	if (!S_ISBLK(stbuf.st_mode) ||
		MAJOR(stbuf.st_rdev) != FLOPPY_MAJOR) {
		fprintf( stderr, "%s is no floppy device!\n", argv[1] );
		exit( 1 );
	}
	if (((MINOR(stbuf.st_rdev) >> 2) & 0x1f) == 0) {
		fprintf( stderr, "%s: unknown format, use device with format spec\n",
				 argv[1] );
		exit( 1 );
	}

	if ((ctrl = open( argv[1], O_RDWR|O_EXCL )) < 0)
		PERROR( "open" );
    if (ioctl( ctrl, FDGETPRM, (int)&param ) < 0)
		PERROR( "ioctl(FDGETPRM)" );

	twist1 = (user_twist1 >= 0) ? user_twist1 : (param.sect + 2) / 5;
	twist2 = (user_twist2 >= 0) ? user_twist2 : twist1;
    printf("%sle-sided, %d tracks, %d sec/track, %d/%d twist."
		   " Total capacity %d kB.\n",
		   param.head ? "Doub" : "Sing",
		   param.track, param.sect, twist1, twist2,
		   param.size >> 1 );
	
    format_disk( argv[1], twist1, twist2 );
    if (verify)
		verify_disk( argv[1] );

    if (close( ctrl ) < 0)
		PERROR( "close" );

	return( 0 );
}
